/* bnx2i.h: Broadcom NetXtreme II iSCSI driver.
 *
 * Copyright (c) 2004, 2005, 2006 Broadcom Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation.
 *
 * Written by: Anil Veerabhadrappa (anilgv@broadcom.com)
 */

#ifndef _BNX2I_H_
#define _BNX2I_H_

#include <linux/version.h>

#include <linux/module.h>
#include <linux/moduleparam.h>

#include <linux/types.h>
#include <linux/list.h>
#include <linux/delay.h>
#include <asm/byteorder.h>
#include <linux/timer.h>
#include <linux/errno.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/dma-mapping.h>
#include <linux/workqueue.h>
#include <linux/mutex.h>
#include <net/tcp.h>
#include <linux/if_vlan.h>

#include <linux/spinlock.h>
#include <linux/kthread.h>
#include <asm/semaphore.h>
#include <linux/bitops.h>

#include <scsi/scsi.h>
#include <scsi/scsi_host.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_eh.h>
#include <scsi/scsi_transport_iscsi.h>
#include <scsi/iscsi_proto.h>

#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <asm/uaccess.h>
#include <asm/io.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25)
#include "../../net/cnic_if.h"
#include "../../net/bnx2.h"
#else
#include <cnic_if.h>
#include <bnx2.h>
#endif
#include "57xx_iscsi_hsi.h"
#include "57xx_iscsi_constants.h"
#include "bnx2i_ioctl.h"

#define BNX2_ISCSI_DRIVER_NAME		"bnx2i"

#define ISCSI_MAX_CONNS_PER_HBA		64
#define ISCSI_MAX_SESS_PER_HBA		ISCSI_MAX_CONNS_PER_HBA
#define ISCSI_MAX_CMDS_PER_SESS		128

#define ISCSI_MAX_BDS_PER_CMD		32

#define MAX_PAGES_PER_CTRL_STRUCT_POOL	8
#define BNX2I_RESERVED_SLOW_PATH_CMD_SLOTS	4

/* 5706/08 hardware has limit on maximum buffer size per BD it can handle */
#define MAX_BD_LENGTH			65535
#define BD_SPLIT_SIZE			32768

/* min, max & default values for SQ/RQ/CQ size, configurable via' modparam */
#define BNX2I_SQ_WQES_MIN 		16
#define BNX2I_SQ_WQES_MAX 		256
#define BNX2I_SQ_WQES_DEFAULT 		128

#define BNX2I_CQ_WQES_MIN 		16
#define BNX2I_CQ_WQES_MAX 		256
#define BNX2I_CQ_WQES_DEFAULT 		128

#define BNX2I_RQ_WQES_MIN 		16
#define BNX2I_RQ_WQES_MAX 		32
#define BNX2I_RQ_WQES_DEFAULT 		16

/* CCELLs per conn */
#define BNX2I_CCELLS_MIN		16
#define BNX2I_CCELLS_MAX		96
#define BNX2I_CCELLS_DEFAULT		64

#define ISCSI_CONN_LOGIN_BUF_SIZE	16384
#define ITT_INVALID_SIGNATURE		0xFFFF

#define ISCSI_CMD_CLEANUP_TIMEOUT	100

#define BNX2I_CONN_CTX_BUF_SIZE		16384

#define BNX2I_SQ_WQE_SIZE		64
#define BNX2I_RQ_WQE_SIZE		256
#define BNX2I_CQE_SIZE			64

#define MB_KERNEL_CTX_SHIFT	8
#define MB_KERNEL_CTX_SIZE	(1 << MB_KERNEL_CTX_SHIFT)

#define CTX_SHIFT		7
#define GET_CID_NUM(cid_addr)	((cid_addr) >> CTX_SHIFT)

#define CTX_OFFSET 		0x10000
#define MAX_CID_CNT		0x4000

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)

#define scsi_sg_count(cmd) ((cmd)->use_sg)
#define scsi_sglist(cmd) ((struct scatterlist *)(cmd)->request_buffer)
#define scsi_bufflen(cmd) ((cmd)->request_bufflen)

static inline void scsi_set_resid(struct scsi_cmnd *cmd, int resid)
{
        cmd->resid = resid;
}

static inline int scsi_get_resid(struct scsi_cmnd *cmd)
{
        return cmd->resid;
}

#define scsi_for_each_sg(cmd, sg, nseg, __i)                    \
        for (__i = 0, sg = scsi_sglist(cmd); __i < (nseg); __i++, (sg)++)

#endif

struct ip_addr {
	u32 ip_type;
#define BNX2I_IP_TYPE_IPV4		0x00
#define BNX2I_IP_TYPE_IPV6		0x01
	union {
		u32 ip_addr4;
		u8 ip_addr6[16];
	} ip;
};


/*
 * following structure defines buffer info for generic pdus
 * such as iSCSI Login, Logout and NOP
 */
struct generic_pdu_resc {
	struct iscsi_hdr pdu_hdr;
	struct bnx2i_cmd *cmd;
	u32 login_itt;
	char *req_buf;
	dma_addr_t req_dma_addr;
	u32 req_buf_size;
	char *req_wr_ptr;
	struct iscsi_hdr resp_hdr;
	char *resp_buf;
	dma_addr_t resp_dma_addr;
	u32 resp_buf_size;
	char *resp_wr_ptr;
	char *req_bd_tbl;
	dma_addr_t req_bd_dma;
	char *resp_bd_tbl;
	dma_addr_t resp_bd_dma;
};



/*
 * structure to track DMA'able memory allocated for command BD tables
 */
struct bd_resc_page {
	struct list_head link;
	u32 max_ptrs;
	u32 num_valid;
	void *page[1];
};


/*
 * IO BD table
 */
struct io_bdt {
	struct list_head link;
	struct iscsi_bd *bd_tbl;
	dma_addr_t bd_tbl_dma;
	struct bnx2i_cmd *cmdp;
	u16 max_bd_cnt;
	u16 bd_valid;
};


/*
 * iSCSI CMD structure
 */
struct bnx2i_cmd {
	struct list_head link;
	int in_use;
	u8 iscsi_opcode;
	u16 rsvd1[3];

	u32 cmd_state;
		#define ISCSI_CMD_STATE_INITIATED		0x001
		#define ISCSI_CMD_STATE_ABORT_PEND		0x002
		#define ISCSI_CMD_STATE_ABORT_COMPL		0x004
		#define ISCSI_CMD_STATE_CLEANUP_PEND		0x008
		#define ISCSI_CMD_STATE_CLEANUP_CMPL		0x010
		#define ISCSI_CMD_STATE_FAILED			0x100
		#define ISCSI_CMD_STATE_TMF_TIMEOUT		0x200
		#define ISCSI_CMD_STATE_COMPLETED		0x800
	int scsi_status_rcvd;

	struct bnx2i_conn *conn;
	struct scsi_cmnd *scsi_cmd;
	u32 tmf_ref_itt;
	struct bnx2i_cmd *tmf_ref_cmd;
	struct scsi_cmnd *tmf_ref_sc;

	struct scatterlist *sg;
	struct io_bdt *bd_tbl;
	dma_addr_t bd_tbl_dma;
	u32 reserved0;

	struct iscsi_cmd_request req;
};


/*
 * TCP port manager
 */
struct tcp_port_mngt {
	int num_required;
	u32 port_tbl_size;
	u32 num_free_ports;
	u32 prod_idx;
	u32 cons_idx;
	u32 max_idx;
	u16 *free_q;
};


/*
 * iSCSI connection structure
 */
struct bnx2i_conn {
	struct list_head link;
	struct bnx2i_sess *sess;
	struct iscsi_cls_conn *cls_conn;

	u32 state;
		#define CONN_STATE_IDLE				0x00
		#define CONN_STATE_XPORT_READY			0x01
		#define CONN_STATE_IN_LOGIN			0x02
		#define CONN_STATE_FFP_STATE			0x04
		#define CONN_STATE_IN_LOGOUT			0x08
		#define CONN_STATE_IN_CLEANUP			0x10
		#define CONN_STATE_XPORT_FREEZE			0x20
		#define CONN_STATE_STOPPED			0x80
	u32 stop_state;
	u32 stage;
	u32 in_shutdown;

	u32 lead_conn;
	u32 conn_cid;

	/*
	 * Following are iSCSI sequencing & operational parameters
	 */
	u32 exp_statsn;
		#define STATSN_UPDATE_SIGNATURE		0xFABCAFE
	u32 header_digest_en;
	u32 data_digest_en;
	u32 max_data_seg_len_xmit;	/* Target */
	u32 max_data_seg_len_recv;	/* Initiator */
	int ifmarker_enable;
	int ofmarker_enable;
	int persist_port;
	char *persist_address;

	u32 iscsi_conn_cid;
		#define BNX2I_CID_RESERVED	0x5AFF
	u32 fw_cid;

	struct timer_list poll_timer;
	spinlock_t lock; /*protects conn structure elements */

	/*
	 * Queue Pair (QP) related structure elements.
	 */
	struct bnx2i_endpoint *ep;

	/*
	 * Buffer for login negotiation process
	 */
	struct generic_pdu_resc gen_pdu;

	u32 nopout_num_scsi_cmds;
	/*
	 * Connection Statistics
	 */
	u64 total_data_octets_sent;
	u64 total_data_octets_rcvd;
	u32 num_login_req_pdus;
	u32 num_login_resp_pdus;
	u32 num_scsi_cmd_pdus;
	u32 num_scsi_resp_pdus;
	u32 num_nopout_pdus;
	u32 num_nopin_pdus;
	u32 num_reject_pdus;
	u32 num_async_pdus;
	u32 num_dataout_pdus;
	u32 num_r2t_pdus;
	u32 num_datain_pdus;
	u32 num_snack_pdus;
	u32 num_text_req_pdus;
	u32 num_text_resp_pdus;
	u32 num_tmf_req_pdus;
	u32 num_tmf_resp_pdus;
	u32 num_logout_req_pdus;
	u32 num_logout_resp_pdus;
};




struct itt_queue {
	void *itt_que_base;
	u32 *itt_que;
	u32 itt_q_prod_idx;
	u32 itt_q_cons_idx;
	u32 itt_q_max_idx;
	u32 itt_q_count;
	struct bnx2i_cmd **itt_cmd;
		#define get_cmnd(sess, itt)	sess->itt_q.itt_cmd[itt]
};


/*
 * iSCSI Session Structure
 */
struct bnx2i_sess {
	struct list_head link;
	struct bnx2i_hba *hba;
	struct Scsi_Host *host;

	u32 state;
		#define BNX2I_SESS_INITIAL		0x01
		#define BNX2I_SESS_IN_FFP		0x02
		#define BNX2I_SESS_NEED_RECOVERY	0x04
		#define BNX2I_SESS_IN_RECOVERY		0x08
		#define BNX2I_SESS_IN_SHUTDOWN		0x10
		#define BNX2I_SESS_IN_LOGOUT		0x40
		#define is_sess_active(_sess)	\
			(((_sess)->state & BNX2I_SESS_IN_FFP))
	unsigned long recovery_state;
		#define ISCSI_SESS_RECOVERY_START	0x01
		#define ISCSI_SESS_RECOVERY_OPEN_ISCSI	0x02
		#define ISCSI_SESS_RECOVERY_COMPLETE 	0x04
		#define ISCSI_SESS_RECOVERY_FAILED	0x08
	unsigned long old_recovery_state;
	int tmf_active;

	spinlock_t lock;	/* protects session structure */

	/* Command abort timer */
	struct timer_list abort_timer;
	/* event wait queue used during error recovery */
	wait_queue_head_t er_wait;

	struct timer_list poll_timer;

	/*
	 * Per session command (task) structure management
	 */
	void *cmd_pages[MAX_PAGES_PER_CTRL_STRUCT_POOL];
	struct list_head pend_cmds;
	u32 num_pend_cmds;
	struct list_head free_cmds;
	int num_free_cmds;
	int allocated_cmds;

	int sq_size;
	struct itt_queue itt_q;
		#define MAX_BD_RESOURCE_PAGES		8

	struct list_head bd_resc_page;
	struct list_head bd_tbl_list;
	struct list_head bd_tbl_active;

	/*
	 * command queue, management parameters
	 */
	struct list_head active_cmds;
	u32 num_active_cmds;

	/*
	 * iSCSI session related sequencing parameters.
	 */
	unsigned int cmdsn;
	unsigned int exp_cmdsn;
	unsigned int max_cmdsn;

	/*
	 * Following pointers are linked to corresponding entry in
	 * operational parameter table associated with this session.
	 * These are to be filled when session becomes operational (FFP).
	 */
	int initial_r2t;
	int max_r2t;
	int imm_data;
	u32 first_burst_len;
	u32 max_burst_len;
	int time2wait;
	int time2retain;
	int pdu_inorder;
	int dataseq_inorder;
	int erl;
	int tgt_prtl_grp;
	char *target_name;

	unsigned char isid[8];
	unsigned short tsih;

	struct bnx2i_conn *lead_conn;
	struct list_head conn_list;
	u32 num_active_conn;
	u32 max_conns;
	u32 conn_id;
	struct iscsi_cmd_request scsi_cmd_wqe;

	/* Driver private statistics */
	u64 violation_notified;
};



/*
 * Per adapter iSCSI CID Queue
 */
struct iscsi_cid_queue {
	void *cid_que_base;
	u32 *cid_que;
	u32 cid_q_prod_idx;
	u32 cid_q_cons_idx;
	u32 cid_q_max_idx;
	u32 cid_free_cnt;
	struct bnx2i_conn **conn_cid_tbl;
};


/*
 * Debug only - Idle counters for MIPS CPUs
 */
struct mips_idle_count {
	u64 cp_idle_count;
	u64 txp_idle_count;
	u64 txp_tdma_count;
	u64 txp_ctx_count;
	u64 txp_hdrq_count;
	u64 tpat_idle_count;
	u64 rxp_idle_count;
	u64 com_idle_count;
};



/*
 * Adapter Data Structure
 */
struct bnx2i_hba {
	struct list_head link;
	struct cnic_dev *cnic;
	struct pci_dev *pci_dev;
	struct net_device *netdev;
	struct class_device class_dev;
	u32 age;
	u32 cnic_dev_type;
		#define CNIC_1GIG_GEN1			0x1	/* 5706/5708 */
		#define CNIC_1GIG_GEN2			0x2	/* 5709	*/
		#define CNIC_10GIG_GEN1			0x4	/* 57710 */
	u32 mail_queue_access;
		#define BNX2I_MQ_KERNEL_MODE		0x0
		#define BNX2I_MQ_KERNEL_BYPASS_MODE	0x1
		#define BNX2I_MQ_BIN_MODE		0x2
	unsigned long  reg_with_cnic;
		#define BNX2I_CNIC_REGISTERED		1

	unsigned long  adapter_state;
		#define ADAPTER_STATE_UP		0
		#define ADAPTER_STATE_GOING_DOWN	1
		#define ADAPTER_STATE_LINK_DOWN		2
		#define ADAPTER_STATE_INIT_FAILED	31
	unsigned int mtu_supported;
		#define BNX2I_MAX_MTU_SUPPORTED		1500

	struct scsi_host_template *scsi_template;
	struct iscsi_transport *iscsi_transport;
		#define BRCM_ISCSI_XPORT_NAME_PREFIX		"bcm570"
		#define BRCM_ISCSI_XPORT_NAME_SIZE_MAX		64
	struct scsi_transport_template *shost_template;

	u32 max_sqes;
	u32 max_rqes;
	u32 max_cqes;
	u32 num_ccell;

	struct list_head active_sess;
	int num_active_sess;
	int ofld_conns_active;

	int max_active_conns;
	struct iscsi_cid_queue cid_que;

	rwlock_t ep_rdwr_lock;
	struct bnx2i_endpoint **ep_ofld_list;
	int ep_ofld_prod_idx;
	int ep_ofld_cons_idx;
	int ep_ofld_max_idx;
	struct bnx2i_endpoint **ep_destroy_list;
	int ep_destroy_prod_idx;
	int ep_destroy_cons_idx;
	int ep_destroy_max_idx;

	/*
	 * BD table to be used with MP (Middle Path requests.
	 */
	char *mp_bd_tbl;
	dma_addr_t mp_bd_dma;
	char *dummy_buffer;
	dma_addr_t dummy_buf_dma;

	spinlock_t lock;	/* protects hba structure access */

	/* Error handling */
	struct timer_list hba_timer;
	wait_queue_head_t eh_wait;
	struct work_struct err_rec_task;
	struct bnx2i_sess **sess_recov_list;
	int sess_recov_prod_idx;
	int sess_recov_cons_idx;
	int sess_recov_max_idx;
	
	unsigned char mac_addr[MAX_ADDR_LEN];

	/*
	 * PCI related info.
	 */
	u16 pci_did;
	u16 pci_vid;
	u16 pci_sdid;
	u16 pci_svid;
	u16 pci_func;
	u16 pci_devno;
	u32 pci_intr_num;

	/*
	 * Following are a bunch of statistics useful during development
	 * and later stage for score boarding.
	 */
	u32 num_wqe_sent;
	u32 num_cqe_rcvd;
	u32 num_intr_claimed;
	u32 link_changed_count;
	u32 ipaddr_changed_count;
	u32 num_sess_opened;
	u32 num_conn_opened;

	struct mips_idle_count mips_idle;

	void *ctx_addr;
	unsigned int ctx_size;
	dma_addr_t ctx_dma_hndl;
	unsigned int ctx_read_cnt;
	unsigned int ctx_ccell_tasks;
	int ictx_poll_mode;
	unsigned int ictx_poll_cid;
	unsigned int ictx_poll_iscsi_cid;
	char *sq_cq_dump;
	char *sq_cq_rdp;
	unsigned int sq_cq_size;
};


/*******************************************************************************
 * 	QP [ SQ / RQ / CQ ] info.
 ******************************************************************************/

/*
 * SQ/RQ/CQ generic structure definition
 */
struct 	sqe {
	u8 sqe_byte[BNX2I_SQ_WQE_SIZE];
};

struct 	rqe {
	u8 rqe_byte[BNX2I_RQ_WQE_SIZE];
};

struct 	cqe {
	u8 cqe_byte[BNX2I_CQE_SIZE];
};


enum {
#if defined(__LITTLE_ENDIAN)
	CNIC_EVENT_COAL_INDEX	= 0x0,
	CNIC_SEND_DOORBELL	= 0x4,
	CNIC_EVENT_CQ_ARM	= 0x7,
	CNIC_RECV_DOORBELL	= 0x8
#elif defined(__BIG_ENDIAN)
	CNIC_EVENT_COAL_INDEX	= 0x2,
	CNIC_SEND_DOORBELL	= 0x6,
	CNIC_EVENT_CQ_ARM	= 0x4,
	CNIC_RECV_DOORBELL	= 0xa
#endif
};



struct qp_info {
	void __iomem *ctx_base;

	struct 	sqe *sq_virt;
	dma_addr_t sq_phys;
	u32 sq_mem_size;

	struct  sqe *sq_prod_qe;
	struct  sqe *sq_cons_qe;
	struct  sqe *sq_last_qe;
	u16 sq_prod_idx;
	u16 sq_cons_idx;
	u32 sqe_left;

	void 	*sq_pgtbl_virt;
	dma_addr_t sq_pgtbl_phys;
	u32 sq_pgtbl_size;	/* set to PAGE_SIZE for 5708 & 5709 */

	struct 	cqe *cq_virt;
	dma_addr_t cq_phys;
	u32 cq_mem_size;

	struct  cqe *cq_prod_qe;
	struct  cqe *cq_cons_qe;
	struct  cqe *cq_last_qe;
	u16 cq_prod_idx;
	u16 cq_cons_idx;
	u32 cqe_left;
	u32 cqe_size;
	u32 cqe_exp_seq_sn;

	void 	*cq_pgtbl_virt;
	dma_addr_t cq_pgtbl_phys;
	u32 cq_pgtbl_size;	/* set to PAGE_SIZE for 5708 & 5709 */

	struct 	rqe *rq_virt;
	dma_addr_t rq_phys;
	u32 rq_mem_size;

	struct  rqe *rq_prod_qe;
	struct  rqe *rq_cons_qe;
	struct  rqe *rq_last_qe;
	u16 rq_prod_idx;
	u16 rq_cons_idx;
	u32 rqe_left;

	void 	*rq_pgtbl_virt;
	dma_addr_t rq_pgtbl_phys;
	u32 rq_pgtbl_size;	/* set to PAGE_SIZE for 5708 & 5709 */
};



/*
 * CID handles
 */
struct ep_handles {
	u32 fw_cid;
	u32 drv_iscsi_cid;
	u16 pg_cid;
	u16 rsvd;
};


/*
 * Endpoint Structure
 */
struct bnx2i_endpoint {
	struct list_head link;
	struct bnx2i_hba *hba;
	struct bnx2i_conn *conn;
	struct bnx2i_sess *sess;
	struct cnic_sock *cm_sk;
	u32 hba_age;
	int in_use;
	u32 state;
		#define EP_STATE_IDLE			0x00000000
		#define EP_STATE_PG_OFLD_START		0x00000001
		#define EP_STATE_PG_OFLD_COMPL		0x00000002
		#define EP_STATE_OFLD_START		0x00000004
		#define EP_STATE_OFLD_COMPL		0x00000008
		#define EP_STATE_CONNECT_START		0x00000010
		#define EP_STATE_CONNECT_COMPL		0x00000020
		#define EP_STATE_ULP_UPDATE_START	0x00000040
		#define EP_STATE_ULP_UPDATE_COMPL	0x00000080
		#define EP_STATE_DISCONN_START		0x00000100
		#define EP_STATE_DISCONN_COMPL		0x00000200
		#define EP_STATE_CLEANUP_START		0x00000400
		#define EP_STATE_CLEANUP_CMPL		0x00000800
		#define EP_STATE_TCP_FIN_RCVD		0x00001000
		#define EP_STATE_TCP_RST_RCVD		0x00002000
		#define EP_STATE_PG_OFLD_FAILED		0x01000000
		#define EP_STATE_ULP_UPDATE_FAILED	0x02000000
		#define EP_STATE_CLEANUP_FAILED		0x04000000
		#define EP_STATE_OFLD_FAILED		0x08000000
		#define EP_STATE_CONNECT_FAILED		0x10000000
		#define EP_STATE_DISCONN_TIMEDOUT	0x20000000

	u16 tcp_port;

	struct qp_info qp;
	struct ep_handles ids;
		#define ep_iscsi_cid	ids.drv_iscsi_cid
		#define ep_cid		ids.fw_cid
		#define ep_pg_cid	ids.pg_cid
	struct timer_list ofld_timer;
	wait_queue_head_t ofld_wait;
};



/*
 * Function Prototypes
 */
extern int bnx2i_reg_device;
void bnx2i_register_device(struct bnx2i_hba *hba);
void bnx2i_check_nx2_dev_busy(void);
void bnx2i_ep_disconnect(uint64_t ep_handle);

void bnx2i_ulp_init(struct cnic_dev *dev);
void bnx2i_ulp_exit(struct cnic_dev *dev);
void bnx2i_start(void *handle);
void bnx2i_stop(void *handle);
void bnx2i_reg_dev_all(void);
void bnx2i_unreg_dev_all(void);
struct bnx2i_hba *get_adapter_list_head(void);

int bnx2i_ioctl_init(void);
void bnx2i_ioctl_cleanup(void);

struct bnx2i_conn *bnx2i_get_conn_from_id(struct bnx2i_hba *hba,
					  u16 iscsi_cid);

int bnx2i_alloc_ep_pool(void);
void bnx2i_release_ep_pool(void);
struct bnx2i_endpoint *bnx2i_ep_ofld_list_next(struct bnx2i_hba *hba);
struct bnx2i_endpoint *bnx2i_ep_destroy_list_next(struct bnx2i_hba *hba);

struct bnx2i_cmd *bnx2i_alloc_cmd(struct bnx2i_sess *sess);
void bnx2i_free_cmd(struct bnx2i_sess *sess, struct bnx2i_cmd *cmd);

struct bnx2i_hba *bnx2i_find_hba_for_cnic(struct cnic_dev *cnic);
struct bnx2i_hba *bnx2i_get_hba_from_template(
	struct scsi_transport_template *scsit);

struct bnx2i_hba *bnx2i_alloc_hba(struct cnic_dev *cnic);
void bnx2i_free_hba(struct bnx2i_hba *hba);

void bnx2i_process_scsi_resp(struct bnx2i_cmd *cmd,
			     struct iscsi_cmd_response *resp_cqe);
int bnx2i_process_nopin(struct bnx2i_conn *conn,
	struct bnx2i_cmd *cmnd, char *data_buf, int data_len);


void bnx2i_update_cmd_sequence(struct bnx2i_sess *sess, u32 expsn, u32 maxsn);

void bnx2i_get_rq_buf(struct bnx2i_conn *conn, char *ptr, int len);
void bnx2i_put_rq_buf(struct bnx2i_conn *conn, int count);

int bnx2i_indicate_login_resp(struct bnx2i_conn *conn);
int bnx2i_indicate_logout_resp(struct bnx2i_conn *conn);
int bnx2i_indicate_async_mesg(struct bnx2i_conn *conn);

void bnx2i_iscsi_unmap_sg_list(struct bnx2i_hba *hba, struct bnx2i_cmd *cmd);

void bnx2i_start_iscsi_hba_shutdown(struct bnx2i_hba *hba);
void bnx2i_iscsi_handle_ip_event(struct bnx2i_hba *hba);
int bnx2i_do_iscsi_sess_recovery(struct bnx2i_sess *sess, int err_code);
void bnx2i_return_failed_command(struct bnx2i_sess *sess,
				 struct bnx2i_cmd *cmd, int err_code);

int bnx2i_get_tcp_port_requirements(void);
void bnx2i_cleanup_tcp_port_mngr(void);
void bnx2i_init_tcp_port_mngr(void);

extern int bnx2i_send_fw_iscsi_init_msg(struct bnx2i_hba *hba);
extern int bnx2i_send_iscsi_login(struct bnx2i_conn *conn,
				  struct bnx2i_cmd *cmnd);
extern int bnx2i_send_iscsi_text(struct bnx2i_conn *conn,
				 struct bnx2i_cmd *cmnd);
extern int bnx2i_send_iscsi_tmf(struct bnx2i_conn *conn,
				struct bnx2i_cmd *cmnd);
extern int bnx2i_send_iscsi_scsicmd(struct bnx2i_conn *conn,
				    struct bnx2i_cmd *cmnd);
extern int bnx2i_send_iscsi_nopout(struct bnx2i_conn *conn,
				   struct bnx2i_cmd *cmnd, u32 ttt,
				   char *datap, int data_len, int unsol);
extern int bnx2i_send_iscsi_logout(struct bnx2i_conn *conn,
				   struct bnx2i_cmd *cmnd);
extern void bnx2i_send_cmd_cleanup_req(struct bnx2i_hba *hba,
				       struct bnx2i_cmd *cmd);
extern void bnx2i_send_conn_ofld_req(struct bnx2i_hba *hba,
				     struct bnx2i_endpoint *ep);
extern void bnx2i_update_iscsi_conn(struct bnx2i_conn *conn);
extern void bnx2i_send_conn_destroy(struct bnx2i_hba *hba,
				    struct bnx2i_endpoint *ep);
extern void bnx2i_recovery_que_add_conn(struct bnx2i_hba *hba,
					struct bnx2i_conn *conn);

extern int bnx2i_alloc_qp_resc(struct bnx2i_hba *hba,
			       struct bnx2i_endpoint *ep);
extern void bnx2i_free_qp_resc(struct bnx2i_hba *hba,
			       struct bnx2i_endpoint *ep);
extern void bnx2i_ep_ofld_timer(unsigned long data);

void bnx2i_map_ep_dbell_regs(struct bnx2i_endpoint *ep);

int bnx2i_register_xport(struct bnx2i_hba *hba);
int bnx2i_deregister_xport(struct bnx2i_hba *hba);

/* Debug related function prototypes */
extern void bnx2i_print_pend_cmd_queue(struct bnx2i_conn *conn);
extern void bnx2i_print_active_cmd_queue(struct bnx2i_conn *conn);
extern void bnx2i_print_xmit_pdu_queue(struct bnx2i_conn *conn);
extern void bnx2i_print_recv_state(struct bnx2i_conn *conn);

#ifndef _SYSFS_INCL_
#define bnx2i_setup_ictx_dump(__hba, __conn)	do { } while (0)
#define  bnx2i_sysfs_setup()			do { } while (0)
#define  bnx2i_sysfs_cleanup()			do { } while (0)
#define  bnx2i_register_sysfs(__hba)		0
#define  bnx2i_unregister_sysfs(__hba)		do { } while (0)
#define bnx2i_init_mips_idle_counters(__hba)	do { } while (0)
#else
extern void bnx2i_setup_ictx_dump(struct bnx2i_hba *hba,
				  struct bnx2i_conn *conn);
extern int bnx2i_sysfs_setup(void);
extern void bnx2i_sysfs_cleanup(void);
extern int bnx2i_register_sysfs(struct bnx2i_hba *hba);
extern void bnx2i_unregister_sysfs(struct bnx2i_hba *hba);
void bnx2i_init_mips_idle_counters(struct bnx2i_hba *hba);
void bnx2i_tcp_port_new_entry(u16 tcp_port);

#endif

#endif
