/* bnx2i.c: Broadcom NetXtreme II iSCSI driver.
 *
 * Copyright (c) 2004, 2005, 2006 Broadcom Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation.
 *
 * Written by: Anil Veerabhadrappa (anilgv@broadcom.com)
 */

#include "bnx2i.h"

static struct list_head adapter_list;
static u32 adapter_count;
int bnx2i_reg_device = 0;

#define DRV_MODULE_NAME		"bnx2i"
#define DRV_MODULE_VERSION	"1.1.2"
#define DRV_MODULE_RELDATE	"Dec 06, 2007"

static char version[] __devinitdata =
		"Broadcom NetXtreme II iSCSI Driver " DRV_MODULE_NAME \
		" v" DRV_MODULE_VERSION " (" DRV_MODULE_RELDATE ")\n";


MODULE_AUTHOR("Anil Veerabhadrappa <anilgv@broadcom.com>");
MODULE_DESCRIPTION("Broadcom NetXtreme II BCM5708 iSCSI Driver");
MODULE_LICENSE("GPL");
MODULE_VERSION(DRV_MODULE_VERSION);

DEFINE_RWLOCK(bnx2i_dev_lock);

unsigned int event_coal_div = 1;
module_param(event_coal_div, int, 0664);
MODULE_PARM_DESC(event_coal_div, "Event Coalescing Divide Factor");

unsigned int bnx2i_nopout_when_cmds_active = 1;
module_param(bnx2i_nopout_when_cmds_active, int, 0664);
MODULE_PARM_DESC(bnx2i_nopout_when_cmds_active,
		"iSCSI NOOP even when connection is not idle");

unsigned int en_tcp_dack = 1;
module_param(en_tcp_dack, int, 0664);
MODULE_PARM_DESC(en_tcp_dack, "Enable TCP Delayed ACK");

unsigned int error_mask1 = 0x00;
module_param(error_mask1, int, 0664);
MODULE_PARM_DESC(error_mask1, "Config FW iSCSI Error Mask #1");

unsigned int error_mask2 = 0x00;
module_param(error_mask2, int, 0664);
MODULE_PARM_DESC(error_mask2, "Config FW iSCSI Error Mask #2");

unsigned int sq_size = BNX2I_SQ_WQES_DEFAULT;
module_param(sq_size, int, 0664);
MODULE_PARM_DESC(sq_size, "Configure SQ size");

unsigned int rq_size = BNX2I_RQ_WQES_DEFAULT;
module_param(rq_size, int, 0664);
MODULE_PARM_DESC(rq_size, "Configure RQ size");

unsigned int cq_size = BNX2I_CQ_WQES_DEFAULT;
module_param(cq_size, int, 0664);
MODULE_PARM_DESC(cq_size, "Configure CQ size");

u64 iscsi_error_mask = 0x00;

extern struct cnic_ulp_ops bnx2i_cnic_cb;
extern spinlock_t bnx2i_resc_lock; /* protects global data structures */
extern struct tcp_port_mngt bnx2i_tcp_port_tbl;

static void bnx2i_unreg_one_device(struct bnx2i_hba *hba) ;


int bnx2i_get_tcp_port_requirements()
{
	struct list_head *list;
	struct list_head *temp;
	struct bnx2i_hba *hba;
	int count = 0;

	read_lock(&bnx2i_dev_lock);
	list_for_each_safe(list, temp, &adapter_list) {
		hba = (struct bnx2i_hba *) list;
		count += hba->max_active_conns;
	}
	read_unlock(&bnx2i_dev_lock);
	return count;
}


struct bnx2i_hba *get_adapter_list_head(void)
{
	struct list_head *list;
	struct bnx2i_hba *hba = NULL;

	if (!adapter_count)
		goto hba_not_found;

	read_lock(&bnx2i_dev_lock);
	list_for_each(list, &adapter_list) {
		hba = (struct bnx2i_hba *) list;
		if (hba && hba->cnic && hba->cnic->cm_select_dev)
			break;
		hba = NULL;
	}
	read_unlock(&bnx2i_dev_lock);
hba_not_found:
	return hba;
}

struct bnx2i_hba *bnx2i_get_hba_from_template(
			struct scsi_transport_template *scsit)
{
	struct list_head *list;
	struct bnx2i_hba *hba = NULL;
	int match_found = 0;

	read_lock(&bnx2i_dev_lock);
	list_for_each(list, &adapter_list) {
		hba = (struct bnx2i_hba *) list;
		if (hba->shost_template == scsit) {
			match_found = 1;
			break;
		}
	}
	read_unlock(&bnx2i_dev_lock);

	if (match_found == 0)
		hba = NULL;
	return hba;
}


struct bnx2i_hba *bnx2i_find_hba_for_cnic(struct cnic_dev *cnic)
{
	struct list_head *list;
	struct list_head *temp;
	struct bnx2i_hba *hba;

	read_lock(&bnx2i_dev_lock);
	list_for_each_safe(list, temp, &adapter_list) {
		hba = (struct bnx2i_hba *) list;

		if (hba->cnic == cnic) {
			read_unlock(&bnx2i_dev_lock);
			return hba;
		}
	}
	read_unlock(&bnx2i_dev_lock);
	return NULL;
}



void bnx2i_start(void *handle)
{
#define BNX2I_INIT_POLL_TIME	(1000 / HZ)
	struct bnx2i_hba *hba = handle;
	int i = HZ;

	bnx2i_send_fw_iscsi_init_msg(hba);
	while (!test_bit(ADAPTER_STATE_UP, &hba->adapter_state) && i--)
		msleep(BNX2I_INIT_POLL_TIME);
}


void bnx2i_stop(void *handle)
{
	struct bnx2i_hba *hba = handle;

	/* check if cleanup happened in GOING_DOWN context */
	clear_bit(ADAPTER_STATE_UP, &hba->adapter_state);
	if (!test_and_clear_bit(ADAPTER_STATE_GOING_DOWN, &hba->adapter_state)) {
		bnx2i_start_iscsi_hba_shutdown(hba);
	}
	
}


void bnx2i_register_device(struct bnx2i_hba *hba)
{
	spin_lock(&hba->lock);	/* called from ep_connect context */
	    
	if (test_bit(ADAPTER_STATE_GOING_DOWN, &hba->adapter_state) ||
	    test_bit(BNX2I_CNIC_REGISTERED, &hba->reg_with_cnic)) {
		goto rel_lock;
	}

	hba->cnic->register_device(hba->cnic, CNIC_ULP_ISCSI, (void *) hba);
	bnx2i_reg_device++;
	set_bit(BNX2I_CNIC_REGISTERED, &hba->reg_with_cnic);
rel_lock:
	spin_unlock(&hba->lock);
}


void bnx2i_reg_dev_all(void)
{
	struct list_head *list;
	struct list_head *temp;

	read_lock(&bnx2i_dev_lock);
	list_for_each_safe(list, temp, &adapter_list)
		bnx2i_register_device((struct bnx2i_hba *) list);
	read_unlock(&bnx2i_dev_lock);
}


static void bnx2i_unreg_one_device(struct bnx2i_hba *hba) 
{
	spin_lock(&hba->lock); /* ep_connect/ep_disconnect() */
	if (hba->ofld_conns_active ||
	    !test_bit(BNX2I_CNIC_REGISTERED, &hba->reg_with_cnic) ||
	    test_bit(ADAPTER_STATE_GOING_DOWN, &hba->adapter_state)) {
		spin_unlock(&hba->lock);
		return;
	}

	hba->cnic->unregister_device(hba->cnic, CNIC_ULP_ISCSI);
	bnx2i_reg_device--;
	/* ep_disconnect could come before NETDEV_DOWN, driver won't
	 * see NETDEV_DOWN as it already unregistered itself.
	 */
	hba->adapter_state = 0;
	clear_bit(BNX2I_CNIC_REGISTERED, &hba->reg_with_cnic);
	spin_unlock(&hba->lock);
}


void bnx2i_unreg_dev_all(void)
{
	struct list_head *list;
	struct list_head *temp;

	read_lock(&bnx2i_dev_lock);
	list_for_each_safe(list, temp, &adapter_list)
		bnx2i_unreg_one_device((struct bnx2i_hba *) list);
	read_unlock(&bnx2i_dev_lock);
}


static int bnx2i_init_one(struct bnx2i_hba *hba, struct cnic_dev *cnic)
{
	int rc;

	hba->netdev = cnic->netdev;
	hba->pci_dev = cnic->pcidev;
	if (hba->pci_dev) {
		pci_dev_get(hba->pci_dev);
		hba->pci_did = hba->pci_dev->device;
		hba->pci_vid = hba->pci_dev->vendor;
		hba->pci_sdid = hba->pci_dev->subsystem_device;
		hba->pci_svid = hba->pci_dev->subsystem_vendor;
		hba->pci_func = PCI_FUNC(hba->pci_dev->devfn);
		hba->pci_devno = PCI_SLOT(hba->pci_dev->devfn);
		hba->pci_intr_num = hba->pci_dev->irq;
	}

	read_lock(&bnx2i_dev_lock);
	if (bnx2i_reg_device &&
		!test_bit(BNX2I_CNIC_REGISTERED, &hba->reg_with_cnic)) {
		spin_lock(&hba->lock);	/* hot plug */
		rc = cnic->register_device(cnic, CNIC_ULP_ISCSI,
					   (void *) hba);
		if (rc)		/* duplicate registration */
			printk(KERN_ERR "bnx2i- dev reg failed\n");
		bnx2i_reg_device++;
		hba->age++;
		set_bit(BNX2I_CNIC_REGISTERED, &hba->reg_with_cnic);
		spin_unlock(&hba->lock);
	}
	read_unlock(&bnx2i_dev_lock);

	if (cnic->netdev)
		memcpy(hba->mac_addr, cnic->netdev->dev_addr, MAX_ADDR_LEN);

	/* Allocate memory & initialize the SCSI/iSCSI host templates */
	rc = bnx2i_register_xport(hba);
	if (rc)
		goto failed_xport_reg;

	/* create 'sysfs' device objects */
	rc = bnx2i_register_sysfs(hba);
	if (rc)
		goto failed_sysfs_reg;

	bnx2i_init_mips_idle_counters(hba);
	bnx2i_tcp_port_tbl.num_required += hba->max_active_conns;

	write_lock(&bnx2i_dev_lock);
	list_add_tail(&hba->link, &adapter_list);
	adapter_count++;
	write_unlock(&bnx2i_dev_lock);

	return 0;

failed_sysfs_reg:
	bnx2i_deregister_xport(hba);
failed_xport_reg:
	cnic->unregister_device(cnic, CNIC_ULP_ISCSI);
	bnx2i_reg_device--;
	clear_bit(BNX2I_CNIC_REGISTERED, &hba->reg_with_cnic);
	return rc;
}


void bnx2i_ulp_init(struct cnic_dev *dev)
{
	struct bnx2i_hba *hba;

	/* Allocate a HBA structure for this device */
	hba = bnx2i_alloc_hba(dev);
	if (!hba) {
		printk(KERN_ERR "init: hba initialization failed\n");
		return;
	}

	/* Get PCI related information and update hba struct members */
	clear_bit(BNX2I_CNIC_REGISTERED, &hba->reg_with_cnic);
	if (bnx2i_init_one(hba, dev))
		printk(KERN_ERR "bnx2i - hba %p init failed\n", hba);
	else
		hba->cnic = dev;
}


void bnx2i_ulp_exit(struct cnic_dev *dev)
{
	struct bnx2i_hba *hba;

	hba = bnx2i_find_hba_for_cnic(dev);
	if (!hba) {
		printk(KERN_INFO "bnx2i_ulp_exit: hba not "
				 "found, dev 0x%p\n", dev);
		return;
	}
	write_lock(&bnx2i_dev_lock);
	list_del_init(&hba->link);
	bnx2i_tcp_port_tbl.num_required -= hba->max_active_conns;
	adapter_count--;

	/* cleanup 'sysfs' devices and classes */
	bnx2i_unregister_sysfs(hba);

	if (test_bit(BNX2I_CNIC_REGISTERED, &hba->reg_with_cnic)) {
		spin_lock(&hba->lock);	/* hot remove */
		hba->cnic->unregister_device(hba->cnic, CNIC_ULP_ISCSI);
		clear_bit(BNX2I_CNIC_REGISTERED, &hba->reg_with_cnic);
		bnx2i_reg_device--;
		spin_unlock(&hba->lock);
	}
	write_unlock(&bnx2i_dev_lock);

	if (hba->pci_dev)
		pci_dev_put(hba->pci_dev);

	bnx2i_deregister_xport(hba);
	bnx2i_free_hba(hba);
}


static int __init bnx2i_mod_init(void)
{
	printk(KERN_INFO "%s", version);

	INIT_LIST_HEAD(&adapter_list);
	adapter_count = 0;

	bnx2i_alloc_ep_pool();
	bnx2i_init_tcp_port_mngr();

	/* create 'sysfs' class object */
	bnx2i_sysfs_setup();

	cnic_register_driver(CNIC_ULP_ISCSI, &bnx2i_cnic_cb);

	bnx2i_ioctl_init();

	return 0;
}


static void __exit bnx2i_mod_exit(void)
{
	struct list_head *list;
	struct list_head *temp;
	struct bnx2i_hba *hba;

	write_lock(&bnx2i_dev_lock);
	list_for_each_safe(list, temp, &adapter_list) {
		list_del_init(list);
		hba = (struct bnx2i_hba *) list;

		/* cleanup 'sysfs' devices and classes */
		bnx2i_unregister_sysfs(hba);

		if (test_bit(BNX2I_CNIC_REGISTERED, &hba->reg_with_cnic)) {
			hba->cnic->unregister_device(hba->cnic, CNIC_ULP_ISCSI);
			clear_bit(BNX2I_CNIC_REGISTERED, &hba->reg_with_cnic);
			bnx2i_reg_device--;
		}

		if (hba->pci_dev)
			pci_dev_put(hba->pci_dev);
		bnx2i_deregister_xport(hba);
		bnx2i_free_hba(hba);
		adapter_count--;
	}
	write_unlock(&bnx2i_dev_lock);

	bnx2i_ioctl_cleanup();
	cnic_unregister_driver(CNIC_ULP_ISCSI);

	bnx2i_sysfs_cleanup();
	bnx2i_cleanup_tcp_port_mngr();
	bnx2i_release_ep_pool();
}

module_init(bnx2i_mod_init);
module_exit(bnx2i_mod_exit);
